from __future__ import annotations

import numpy as np
import numpy.typing as npt

from ._ipd_dens_funcs import DENSITY_FUNCS
from ._ipd_model import InterplanetaryDustModel
from .ipd_models import model_registry

DEFAULT_EARTH_POS = (1, 0, 0)


def tabulate_density(
    grid: npt.NDArray[np.float64] | list[npt.NDArray[np.float64]],
    model: str | InterplanetaryDustModel = "DIRBE",
    earth_position: tuple[float, float, float]
    | npt.NDArray[np.float64] = DEFAULT_EARTH_POS,
) -> npt.NDArray[np.float64]:
    """Returns the tabulated densities of the Interplanetary Dust components.

    Parameters
    ----------
    grid
        A cartesian mesh grid (x, y, z) created with `np.meshgrid` for which to
        tabulate the Interplanetary dust components.
    model
        The model who's Interplanetary Dust components to tabulate.
    earth_position
        The position of the Earth.

    Returns
    -------
    density_grid
        The tabulate densities of the Interplanetary Dust components.
    """

    if not isinstance(model, InterplanetaryDustModel):
        model = model_registry.get_model(model)

    if not isinstance(grid, np.ndarray):
        grid = np.asarray(grid)

    earth_position = np.reshape(earth_position, (3, 1, 1, 1))

    density_grid = np.zeros((model.n_comps, *grid.shape[1:]))
    for idx, comp in enumerate(model.comps.values()):
        comp.X_0 = np.reshape(comp.X_0, (3, 1, 1, 1))
        density_grid[idx] = DENSITY_FUNCS[type(comp)](
            X_helio=grid,
            X_earth=earth_position,
        )
        comp.X_0 = np.reshape(comp.X_0, (3, 1))

    return density_grid



from . import ipd_comps, source_params
from ._ipd_model import model_registry

model_registry.register_model(
    name="dirbe",
    comps=ipd_comps.DIRBE,
    spectrum=source_params.SPECTRUM_DIRBE,
    emissivities=source_params.EMISSIVITY_DIRBE,
    albedos=source_params.ALBEDO_DIRBE,
    solar_irradiance=source_params.SOLAR_IRRADIANCE_DIRBE,
    phase_coefficients=source_params.PHASE_FUNCTION_DIRBE,
)

model_registry.register_model(
    name="planck13",
    comps=ipd_comps.DIRBE,
    spectrum=source_params.SPECTRUM_PLANCK,
    emissivities=source_params.EMISSIVITY_PLANCK_13,
)

model_registry.register_model(
    name="planck15",
    comps=ipd_comps.PLANCK,
    spectrum=source_params.SPECTRUM_PLANCK,
    emissivities=source_params.EMISSIVITY_PLANCK_15,
)

model_registry.register_model(
    name="planck18",
    comps=ipd_comps.PLANCK,
    spectrum=source_params.SPECTRUM_PLANCK,
    emissivities=source_params.EMISSIVITY_PLANCK_18,
)

model_registry.register_model(
    name="odegard",
    comps=ipd_comps.PLANCK,
    spectrum=source_params.SPECTRUM_PLANCK,
    emissivities=source_params.EMISSIVITY_ODEGARD,
)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Mapping, Sequence

import astropy.units as u

from ._ipd_comps import Component, ComponentLabel
from .source_params import DELTA_DIRBE, T_0_DIRBE


@dataclass
class InterplanetaryDustModel:
    """Container for interplanetary dust components and model parameters."""

    name: str
    comps: Mapping[ComponentLabel, Component]
    spectrum: u.Quantity[u.Hz] | u.Quantity[u.m]
    emissivities: Mapping[ComponentLabel, Sequence[float]]
    albedos: Mapping[ComponentLabel, Sequence[float]] | None = None
    solar_irradiance: u.Quantity[u.MJy / u.sr] | None = None
    phase_coefficients: Sequence[Sequence[float]] | None = None
    T_0: float = T_0_DIRBE
    delta: float = DELTA_DIRBE

    @property
    def n_comps(self) -> int:
        return len(self.comps)

    def __repr__(self) -> str:
        comp_names = [comp_label.value for comp_label in self.comps]
        repr = f"{type(self).__name__}( \n"
        repr += f"   name: {self.name!r},\n"
        repr += "   components: (\n"
        for name in comp_names:
            repr += f"      {name!r},\n"
        repr += "   ),\n"
        repr += "   thermal: True,\n"
        repr += f"   scattering: {self.albedos is not None},\n"
        repr += ")"

        return repr


@dataclass
class InterplanetaryDustModelRegistry:
    """Container for registered models."""

    _registry: dict[str, InterplanetaryDustModel] = field(
        init=False, repr=False, default_factory=dict
    )

    @property
    def models(self) -> list[str]:
        return list(self._registry.keys())

    def register_model(
        self,
        name: str,
        comps: Mapping[ComponentLabel, Component],
        spectrum: u.Quantity[u.Hz] | u.Quantity[u.m],
        emissivities: Mapping[ComponentLabel, Sequence[float]],
        albedos: Mapping[ComponentLabel, Sequence[float]] | None = None,
        solar_irradiance: u.Quantity[u.MJy / u.sr] | None = None,
        phase_coefficients: Sequence[Sequence[float]] | None = None,
        T_0: float = T_0_DIRBE,
        delta: float = DELTA_DIRBE,
    ) -> None:
        if (name := name.lower()) in self._registry:
            raise ValueError(f"a model by the name {name!s} is already registered.")

        if albedos is not None and solar_irradiance is None:
            raise ValueError(
                "albedos are provided but no solar irradiance is provided."
            )
        self._registry[name] = InterplanetaryDustModel(
            name=name,
            comps=comps,
            spectrum=spectrum,
            emissivities=emissivities,
            T_0=T_0,
            delta=delta,
            albedos=albedos,
            solar_irradiance=solar_irradiance,
            phase_coefficients=phase_coefficients,
        )

    def get_model(self, name: str) -> InterplanetaryDustModel:
        if (name := name.lower()) not in self._registry:
            raise ValueError(
                f"{name!r} is not a registered Interplanetary Dust model. "
                f"Avaliable models are: {', '.join(self._registry)}."
            )

        return self._registry[name]


model_registry = InterplanetaryDustModelRegistry()



from __future__ import annotations

from dataclasses import dataclass
from functools import partial

import astropy.units as u
import numpy as np
import numpy.typing as npt
from scipy.interpolate import interp1d

from ._constants import SPECIFIC_INTENSITY_UNITS
from ._ipd_model import InterplanetaryDustModel
from ._types import FrequencyOrWavelength


@dataclass
class InterpolatedSourceParameters:
    emissivities: npt.NDArray[np.float64]
    albedos: npt.NDArray[np.float64]
    phase_coefficients: tuple[float, ...]
    solar_irradiance: float


def interpolate_source_parameters(
    model: InterplanetaryDustModel,
    freq: FrequencyOrWavelength,
    weights: npt.NDArray[np.float64] | None = None,
) -> InterpolatedSourceParameters:

    interpolator = partial(interp1d, x=model.spectrum, fill_value="extrapolate")
    emissivities = np.asarray(
        [
            interpolator(y=model.emissivities[comp_label])(freq)
            for comp_label in model.comps.keys()
        ]
    )
    if model.albedos is not None:
        albedos = np.asarray(
            [
                interpolator(y=model.albedos[comp_label])(freq)
                for comp_label in model.comps.keys()
            ]
        )
    else:
        albedos = np.zeros_like(emissivities)

    if model.phase_coefficients is not None:  # double check this
        phase_coefficients = interpolator(y=np.asarray(model.phase_coefficients))(freq)

    else:

        phase_coefficients = np.repeat(np.zeros((3, 1)), repeats=freq.size, axis=-1)

    if model.solar_irradiance is not None:
        solar_irradiance = interpolator(y=model.solar_irradiance)(freq)
        solar_irradiance = (
            (solar_irradiance * model.solar_irradiance.unit)
            .to(SPECIFIC_INTENSITY_UNITS, equivalencies=u.spectral())
            .value
        )
    else:
        solar_irradiance = 0

    if weights is not None:
        emissivities = np.trapz(weights * emissivities, freq.value, axis=-1)
        albedos = np.trapz(weights * albedos, freq.value, axis=-1)
        phase_coefficients = np.trapz(weights * phase_coefficients, freq.value, axis=-1)
        solar_irradiance = np.trapz(weights * solar_irradiance, freq.value, axis=-1)

    return InterpolatedSourceParameters(
        emissivities=emissivities,
        albedos=albedos,
        phase_coefficients=tuple(phase_coefficients),
        solar_irradiance=solar_irradiance,
    )



from __future__ import annotations

from typing import Sequence

import astropy.units as u
import healpy as hp
import numpy as np
import numpy.typing as npt


def get_unit_vectors_from_pixels(
    coord_in: str, pixels: Sequence[int] | npt.NDArray[np.int64], nside: int
) -> npt.NDArray[np.float64]:
    """Returns ecliptic unit vectors from HEALPix pixels representing some pointing."""

    unit_vectors = np.asarray(hp.pix2vec(nside, pixels))

    return np.asarray(hp.Rotator(coord=[coord_in, "E"])(unit_vectors))


def get_unit_vectors_from_ang(
    coord_in: str,
    phi: u.Quantity[u.rad] | u.Quantity[u.deg],
    theta: u.Quantity[u.rad] | u.Quantity[u.deg],
    lonlat: bool = False,
) -> npt.NDArray[np.float64]:
    """Returns ecliptic unit vectors from sky angles representing some pointing."""

    unit_vectors = np.asarray(hp.ang2vec(theta, phi, lonlat=lonlat)).transpose()

    return np.asarray(hp.Rotator(coord=[coord_in, "E"])(unit_vectors))



from __future__ import annotations

from functools import lru_cache

import numba
import numpy as np
import numpy.typing as npt
from numpy.typing import NDArray

from ._constants import c, h, k_B
from ._types import FloatOrNDArray


@numba.njit(cache=True, fastmath=True)
def get_blackbody_emission(freq: FloatOrNDArray, T: FloatOrNDArray) -> FloatOrNDArray:
    """Returns the blackbody emission given a frequency.

    Parameters
    ----------
    freq
        Frequency [Hz].
    T
        Temperature of the blackbody [K].

    Returns
    -------
        Blackbody emission [W / m^2 Hz sr].
    """
    term1 = (2 * h * freq**3) / c**2
    term2 = np.expm1(((h * freq) / (k_B * T)))

    return term1 / term2


@numba.njit(cache=True, fastmath=True)
def get_bandpass_integrated_blackbody_emission(
    freq: npt.NDArray[np.float64],
    weights: npt.NDArray[np.float64],
    T: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """Returns the bandpass integrated blackbody emission.

    Parameters
    ----------
    freq
        Bandpass frequencies [Hz].
    weights
        Bandpass weights.
    T
        Temperature of the blackbody [K].

    Returns
    -------
        Bandpass integrated blackbody emission [W / m^2 Hz sr].
    """

    emission = np.zeros_like(T)
    delta_freq = np.diff(freq)

    for idx in range(1, len(freq)):
        current = get_blackbody_emission(freq[idx], T) * weights[idx]
        previous = get_blackbody_emission(freq[idx - 1], T) * weights[idx - 1]
        emission += (current + previous) * delta_freq[idx - 1]

    return emission * 0.5


def get_dust_grain_temperature(
    R: FloatOrNDArray, T_0: float, delta: float
) -> FloatOrNDArray:
    """Returns the dust grain temperature given a radial distance from the Sun.

    Parameters
    ----------
    R
        Radial distance from the sun in ecliptic heliocentric coordinates [AU / 1AU].
    T_0
        Temperature of dust grains located 1 AU from the Sun [K].
    delta
        Powerlaw index.

    Returns
    -------
        Dust grain temperature [K].
    """

    return T_0 * R**-delta


def get_scattering_angle(
    R_los: float | npt.NDArray[np.float64],
    R_helio: npt.NDArray[np.float64],
    X_los: npt.NDArray[np.float64],
    X_helio: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    Returns the scattering angle between the Sun and a point along the
    line of sight.

    Parameters
    ----------
    R_los
        Distance from observer to a point along the line of sight.
    R_helio
        Distance from the Sun to a point along the line of sight.
    X_los
        vector pointing from the observer to a point along the line of sight.
    X_helio
        Heliocentric position of point along the line of sight.

    Returns
    -------
        Scattering angle.
    """
    cos_theta = (X_los * X_helio).sum(axis=0) / (R_los * R_helio)
    cos_theta = np.clip(cos_theta, -1, 1)

    return np.arccos(-cos_theta)


def get_phase_function(
    Theta: NDArray[np.float64], C: tuple[float, ...]
) -> NDArray[np.float64]:
    """Returns the phase function.

    Parameters
    ----------
    Theta
        Scattering angle.
    coeffs
        Phase function parameters.

    Returns
    -------
        The Phase funciton.
    """

    phase_normalization = _get_phase_normalization(C)

    return phase_normalization * (C[0] + C[1] * Theta + np.exp(C[2] * Theta))


@lru_cache
def _get_phase_normalization(C: tuple[float, ...]) -> float:
    """Returns the analyitcal integral for the phase normalization factor N."""

    int_term1 = 2 * np.pi
    int_term2 = 2 * C[0]
    int_term3 = np.pi * C[1]
    int_term4 = (np.exp(C[2] * np.pi) + 1) / (C[2] ** 2 + 1)

    return 1 / (int_term1 * (int_term2 + int_term3 + int_term4))



import pkg_resources

from . import ipd_comps, source_params
from ._contour import tabulate_density
from .ipd_models import model_registry
from .zodipy import Zodipy

try:
    __version__ = pkg_resources.get_distribution(__name__).version
except pkg_resources.DistributionNotFound:
    ...

__all__ = (
    "Zodipy",
    "model_registry",
    "ipd_comps",
    "source_params",
    "tabulate_density",
)



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass, field
from enum import Enum

import numpy as np
import numpy.typing as npt


@dataclass
class Component(ABC):
    """Base class for storing common model parameters for zodiacal components.

    Parameters
    ----------
    x_0
        x-offset from the Sun in heliocentric ecliptic coordinates [AU].
    y_0
        y-offset from the Sun in heliocentric ecliptic coordinates [AU].
    z_0
        z-offset from the Sun in heliocentric ecliptic coordinates [AU].
    i
        Inclination with respect to the ecliptic plane [deg].
    Omega
        Ascending node [deg].
    """

    x_0: float = field(repr=False)
    y_0: float = field(repr=False)
    z_0: float = field(repr=False)
    i: float = field(repr=False)
    Omega: float = field(repr=False)

    X_0: npt.NDArray[np.float64] = field(init=False)
    sin_i_rad: float = field(init=False)
    cos_i_rad: float = field(init=False)
    sin_Omega_rad: float = field(init=False)
    cos_Omega_rad: float = field(init=False)

    def __post_init__(self) -> None:
        self.X_0 = np.array([self.x_0, self.y_0, self.z_0]).reshape(3, 1, 1)
        self.sin_i_rad = np.sin(np.radians(self.i))
        self.cos_i_rad = np.cos(np.radians(self.i))
        self.sin_Omega_rad = np.sin(np.radians(self.Omega))
        self.cos_Omega_rad = np.cos(np.radians(self.Omega))


@dataclass
class Cloud(Component):
    """DIRBE diffuse cloud.

    Parameters
    ----------
    n_0
        Density at 1 AU.
    alpha
        Radial power-law exponent.
    beta
       Vertical shape parameter.
    gamma
        Vertical power-law exponent.
    mu
        Widening parameter for the modified fan.
    """

    n_0: float
    alpha: float
    beta: float
    gamma: float
    mu: float


@dataclass
class Band(Component):
    """DIRBE asteroidal dust band.

    Parameters
    ----------
    n_0
        Density at 3 AU.
    delta_zeta
        Shape parameter [deg].
    v
        Shape parameter.
    p
        Shape parameter.
    delta_r
        Inner radial cutoff.
    """

    n_0: float
    delta_zeta: float
    v: float
    p: float
    delta_r: float
    delta_zeta_rad: float = field(init=False)

    def __post_init__(self) -> None:
        super().__post_init__()
        self.delta_zeta_rad = np.radians(self.delta_zeta)


@dataclass
class Ring(Component):
    """DIRBE circum-solar ring (excluding the Earth-trailing Feature).

    Parameters
    ----------
    n_0
        Density at 1 AU.
    R
        Radius of the peak density.
    sigma_r
        Radial dispersion.
    sigma_z
        Vertical dispersion.
    """

    n_0: float
    R: float
    sigma_r: float
    sigma_z: float


@dataclass
class Feature(Component):
    """DIRBE Earth-trailing Feature.

    Parameters
    ----------
    n_0
        Density at 1 AU.
    R
        Radius of the peak density.
    sigma_r
        Radial dispersion.
    sigma_z
        Vertical dispersion.
    theta
        Longitude with respect to Earth [deg].
    sigma_theta
        Longitude dispersion [deg].
    """

    n_0: float
    R: float
    sigma_r: float
    sigma_z: float
    theta: float
    sigma_theta: float
    theta_rad: float = field(init=False)
    sigma_theta_rad: float = field(init=False)

    def __post_init__(self) -> None:
        super().__post_init__()
        self.theta_rad = np.radians(self.theta)
        self.sigma_theta_rad = np.radians(self.sigma_theta)


class ComponentLabel(Enum):
    """Labels representing the components in the DIRBE model."""

    CLOUD = "cloud"
    BAND1 = "band1"
    BAND2 = "band2"
    BAND3 = "band3"
    RING = "ring"
    FEATURE = "feature"



from typing import Tuple, Union

import astropy.units as u
import healpy as hp
import numpy as np
import numpy.typing as npt

from ._ipd_model import InterplanetaryDustModel
from ._types import FrequencyOrWavelength, Pixels, SkyAngles


@u.quantity_input(equivalencies=u.spectral())
def validate_frequency_in_model_range(
    freq: FrequencyOrWavelength, model: InterplanetaryDustModel
) -> None:
    """Validate user inputted frequency."""

    freq_in_spectrum_units = freq.to(model.spectrum.unit, equivalencies=u.spectral())

    lower_freq_range = model.spectrum.min()
    upper_freq_range = model.spectrum.max()

    if freq_in_spectrum_units.isscalar:
        freq_is_in_range = (
            lower_freq_range <= freq_in_spectrum_units <= upper_freq_range
        )
    else:
        freq_is_in_range = all(
            lower_freq_range.value <= nu <= upper_freq_range.value and nu
            for nu in freq_in_spectrum_units.value
        )

    if not freq_is_in_range:
        raise ValueError(
            f"Model: {model.name} is only valid in the [{lower_freq_range},"
            f" {upper_freq_range}] range."
        )


@u.quantity_input
def validate_and_normalize_weights(
    freq: FrequencyOrWavelength, weights: Union[u.Quantity[u.MJy / u.sr], None]
) -> Union[npt.NDArray[np.float64], None]:
    """Validate user inputted weights."""

    if weights is not None:
        weights = weights.to(u.MJy / u.sr)
        return (weights / np.trapz(weights, freq)).value

    if weights is None and not freq.isscalar:
        print("Warning: weights not provided, assuming uniform weights.")
        return np.ones_like(freq).value / len(freq)

    return None


@u.quantity_input(theta=[u.deg, u.rad], phi=[u.deg, u.rad])
def validate_ang(
    theta: SkyAngles, phi: SkyAngles, lonlat: bool
) -> Tuple[SkyAngles, SkyAngles]:
    """Validate user inputted sky angles."""

    theta = theta.to(u.deg) if lonlat else theta.to(u.rad)
    phi = phi.to(u.deg) if lonlat else phi.to(u.rad)

    if theta.isscalar:
        theta = u.Quantity([theta])
    if phi.isscalar:
        phi = u.Quantity([phi])

    return theta, phi


def validate_pixels(pixels: Pixels, nside: int) -> Pixels:
    """Validate user inputted pixels."""

    if (np.max(pixels) > hp.nside2npix(nside)) or (np.min(pixels) < 0):
        raise ValueError("invalid pixel number given nside")

    if np.ndim(pixels) == 0:
        pixels = np.expand_dims(pixels, axis=0)

    return pixels



from __future__ import annotations

import multiprocessing
import platform
from dataclasses import asdict
from functools import partial
from typing import Literal

import astropy.units as u
import healpy as hp
import numpy as np
import numpy.typing as npt
import quadpy
from astropy.coordinates import solar_system_ephemeris
from astropy.time import Time

from ._constants import SPECIFIC_INTENSITY_UNITS
from ._interp import interpolate_source_parameters
from ._ipd_dens_funcs import PartialComputeDensityFunc, construct_density_partials
from ._line_of_sight import get_line_of_sight_endpoints
from ._sky_coords import DISTANCE_TO_JUPITER, get_obs_and_earth_positions
from ._source_funcs import (
    get_bandpass_integrated_blackbody_emission,
    get_blackbody_emission,
    get_dust_grain_temperature,
    get_phase_function,
    get_scattering_angle,
)
from ._types import FrequencyOrWavelength, Pixels, SkyAngles
from ._unit_vectors import get_unit_vectors_from_ang, get_unit_vectors_from_pixels
from ._validators import (
    validate_and_normalize_weights,
    validate_ang,
    validate_frequency_in_model_range,
    validate_pixels,
)
from .ipd_models import model_registry

PLATFORM = platform.system().lower()
SYS_PROC_START_METHOD = "fork" if "windows" not in PLATFORM else None


class Zodipy:
    """Interface for simulating zodiacal emission.

    Sets up the simulation configuration and provides methods for simulating the zodiacal
    emission that a solar system observer is predicted to see given an interplanetary dust
    model.

    Attributes:
        model (str): Name of the interplanetary dust model to use in the simulations.
            Defaults to DIRBE.
        extrapolate (bool): If `True` all spectral quantities in the selected model are
            linearly extrapolated to the requested frequency/wavelength. If `False`, an
            exception is raised on requested frequencies/wavelengths outside of the
            valid model range. Default is `False`.
        parallel (bool): If `True`, input pointing sequences will be split among all
            available cores on the machine, and the emission will be computed in parallel.
            This is useful for large simulations. Default is `True`.
        n_proc (int): Number of cores to use when parallel computation. Defaults is None,
            which will use all available cores.
        solar_cut (u.Quantity[u.deg]): Cutoff angle from the sun in degrees. The emission
            for all the pointing with angular distance between the sun smaller than
            `solar_cutoff` are set to `np.nan`. This is due to the model singularity of the
            diffuse cloud component at the heliocentric origin. Such a cutoff may be
            useful when using simulated pointing, but actual scanning strategies are
            unlikely to look directly at the sun. This feature is turned of by setting this
            argument to `None`. Defaults to 5 degrees.
        solar_cut_fill_value (float): Fill value for the masked solar cut pointing.
            Defaults to `np.nan`.
        gauss_quad_degree (int): Order of the Gaussian-Legendre quadrature used to evaluate
            the brightness integral. Default is 100 points.
        los_dist_cut (u.Quantity[u.AU]): Radial distance from the Sun at which all line of
            sights are truncated. Defaults to 5.2 AU which is the distance to Jupiter.
        ephemeris (str): Ephemeris used to compute the positions of the observer and the
            Earth. Defaults to 'de432s' which requires downloading (and caching) a ~10MB
            file. For more information on available ephemeridis, please visit
            https://docs.astropy.org/en/stable/coordinates/solarsystem.html

    """

    def __init__(
        self,
        model: str = "dirbe",
        extrapolate: bool = False,
        parallel: bool = True,
        n_proc: int | None = None,
        solar_cut: u.Quantity[u.deg] | None = None,
        solar_cut_fill_value: float = np.nan,
        gauss_quad_degree: int = 100,
        los_dist_cut: u.Quantity[u.AU] = DISTANCE_TO_JUPITER,
        ephemeris: str = "de432s",
    ) -> None:

        self.model = model
        self.extrapolate = extrapolate
        self.parallel = parallel
        self.n_proc = n_proc
        self.solar_cut = solar_cut.to(u.rad) if solar_cut is not None else solar_cut
        self.solar_cut_fill_value = solar_cut_fill_value
        self.gauss_quad_degree = gauss_quad_degree
        self.los_dist_cut = los_dist_cut
        self.ephemeris = ephemeris

        self.ipd_model = model_registry.get_model(model)
        self._integration_scheme = quadpy.c1.gauss_legendre(self.gauss_quad_degree)

    @property
    def supported_observers(self) -> list[str]:
        """Returns a list of available observers given an ephemeris."""

        return list(solar_system_ephemeris.bodies) + ["semb-l2"]

    def get_emission_ang(
        self,
        freq: FrequencyOrWavelength,
        theta: SkyAngles,
        phi: SkyAngles,
        obs_time: Time,
        obs: str = "earth",
        obs_pos: u.Quantity[u.AU] | None = None,
        weights: u.Quantity[u.MJy / u.sr] | None = None,
        lonlat: bool = False,
        return_comps: bool = False,
        coord_in: Literal["E", "G", "C"] = "E",
    ) -> u.Quantity[u.MJy / u.sr]:
        """Returns the simulated zodiacal emission given angles on the sky.

        The pointing, for which to compute the emission, is specified in form of angles on
        the sky given by `theta` and `phi`.

        Args:
            freq: Delta frequency/wavelength or a sequence of frequencies corresponding to
                a bandpass over which to evaluate the zodiacal emission.
            theta: Angular co-latitude coordinate of a point, or a sequence of points, on
                the celestial sphere. Must be in the range [0, π] rad. Units must be either
                radians or degrees.
            phi: Angular longitude coordinate of a point, or a sequence of points, on the
                celestial sphere. Must be in the range [0, 2π] rad. Units must be either
                radians or degrees.
            obs_time: Time of observation.
            obs: Name of the Solar System observer. A list of all support observers (for a
                given ephemeridis) is specified in `supported_observers` attribute of the
                `Zodipy` instance. Defaults to 'earth'.
            obs_pos: The heliocentric ecliptic cartesian position of the observer in AU.
                Overrides the `obs` argument. Default is None.
            weights: Bandpass weights corresponding the the frequencies in `freq`. The
                weights are assumed to be in units of MJy/sr.
            lonlat: If True, input angles (`theta`, `phi`) are assumed to be longitude and
                latitude, otherwise, they are co-latitude and longitude.
            return_comps: If True, the emission is returned component-wise. Defaults to False.
            coord_in: Coordinate frame of the input pointing. Assumes 'E' (ecliptic
                coordinates) by default.

        Returns:
            emission: Simulated zodiacal emission in units of 'MJy/sr'.

        """

        theta, phi = validate_ang(theta=theta, phi=phi, lonlat=lonlat)

        unique_angles, indicies = np.unique(
            np.asarray([theta, phi]), return_inverse=True, axis=1
        )
        unit_vectors = get_unit_vectors_from_ang(
            coord_in=coord_in,
            theta=unique_angles[0],
            phi=unique_angles[1],
            lonlat=lonlat,
        )

        return self._compute_emission(
            freq=freq,
            weights=weights,
            obs=obs,
            obs_time=obs_time,
            obs_pos=obs_pos,
            unit_vectors=unit_vectors,
            indicies=indicies,
            return_comps=return_comps,
        )

    def get_emission_pix(
        self,
        freq: FrequencyOrWavelength,
        pixels: Pixels,
        nside: int,
        obs_time: Time,
        obs: str = "earth",
        obs_pos: u.Quantity[u.AU] | None = None,
        weights: u.Quantity[u.MJy / u.sr] | None = None,
        return_comps: bool = False,
        coord_in: Literal["E", "G", "C"] = "E",
    ) -> u.Quantity[u.MJy / u.sr]:
        """Returns the simulated zodiacal emission given pixel numbers.

        The pixel numbers represent the pixel indicies on a HEALPix grid with resolution
        given by `nside`.

        Args:
            freq: Delta frequency/wavelength or a sequence of frequencies corresponding to
                a bandpass over which to evaluate the zodiacal emission.
            pixels: HEALPix pixel indicies representing points on the celestial sphere.
            nside: HEALPix resolution parameter of the pixels and the binned map.
            obs_time: Time of observation.
            obs: Name of the Solar System observer. A list of all support observers (for a
                given ephemeridis) is specified in `supported_observers` attribute of the
                `Zodipy` instance. Defaults to 'earth'.
            obs_pos: The heliocentric ecliptic cartesian position of the observer in AU.
                Overrides the `obs` argument. Default is None.
            weights: Bandpass weights corresponding the the frequencies in `freq`. The
                weights are assumed to be in units of MJy/sr.
            return_comps: If True, the emission is returned component-wise. Defaults to False.
            coord_in: Coordinate frame of the input pointing. Assumes 'E' (ecliptic
                coordinates) by default.

        Returns:
            emission: Simulated zodiacal emission in units of 'MJy/sr'.

        """

        pixels = validate_pixels(pixels=pixels, nside=nside)

        unique_pixels, indicies = np.unique(pixels, return_inverse=True)
        unit_vectors = get_unit_vectors_from_pixels(
            coord_in=coord_in,
            pixels=unique_pixels,
            nside=nside,
        )

        return self._compute_emission(
            freq=freq,
            weights=weights,
            obs=obs,
            obs_time=obs_time,
            obs_pos=obs_pos,
            unit_vectors=unit_vectors,
            indicies=indicies,
            pixels=unique_pixels,
            nside=nside,
            return_comps=return_comps,
        )

    def get_binned_emission_ang(
        self,
        freq: FrequencyOrWavelength,
        theta: SkyAngles,
        phi: SkyAngles,
        nside: int,
        obs_time: Time,
        obs: str = "earth",
        obs_pos: u.Quantity[u.AU] | None = None,
        weights: u.Quantity[u.MJy / u.sr] | None = None,
        lonlat: bool = False,
        return_comps: bool = False,
        coord_in: Literal["E", "G", "C"] = "E",
    ) -> u.Quantity[u.MJy / u.sr]:
        """Returns the simulated binned zodiacal emission given angles on the sky.

        The pointing, for which to compute the emission, is specified in form of angles on
        the sky given by `theta` and `phi`. The emission is binned to a HEALPix map with
        resolution given by `nside`.

        Args:
            freq: Delta frequency/wavelength or a sequence of frequencies corresponding to
                a bandpass over which to evaluate the zodiacal emission.
            theta: Angular co-latitude coordinate of a point, or a sequence of points, on
                the celestial sphere. Must be in the range [0, π] rad. Units must be either
                radians or degrees.
            phi: Angular longitude coordinate of a point, or a sequence of points, on the
                celestial sphere. Must be in the range [0, 2π] rad. Units must be either
                radians or degrees.
            nside: HEALPix resolution parameter of the binned map.
            obs_time: Time of observation.
            obs: Name of the Solar System observer. A list of all support observers (for a
                given ephemeridis) is specified in `supported_observers` attribute of the
                `Zodipy` instance. Defaults to 'earth'.
            obs_pos: The heliocentric ecliptic cartesian position of the observer in AU.
                Overrides the `obs` argument. Default is None.
            weights: Bandpass weights corresponding the the frequencies in `freq`. The
                weights are assumed to be in units of MJy/sr.
            lonlat: If True, input angles `theta`, `phi` are assumed to be longitude and
                latitude, otherwise, they are co-latitude and longitude.
            return_comps: If True, the emission is returned component-wise. Defaults to False.
            coord_in: Coordinate frame of the input pointing. Assumes 'E' (ecliptic
                coordinates) by default.

        Returns:
            emission: Simulated zodiacal emission in units of 'MJy/sr'.

        """

        theta, phi = validate_ang(theta=theta, phi=phi, lonlat=lonlat)

        unique_angles, counts = np.unique(
            np.asarray([theta, phi]), return_counts=True, axis=1
        )
        unique_pixels = hp.ang2pix(nside, *unique_angles, lonlat=lonlat)
        unit_vectors = get_unit_vectors_from_ang(
            coord_in=coord_in,
            theta=unique_angles[0],
            phi=unique_angles[1],
            lonlat=lonlat,
        )

        return self._compute_emission(
            freq=freq,
            weights=weights,
            obs=obs,
            obs_time=obs_time,
            obs_pos=obs_pos,
            unit_vectors=unit_vectors,
            indicies=counts,
            binned=True,
            pixels=unique_pixels,
            nside=nside,
            return_comps=return_comps,
        )

    def get_binned_emission_pix(
        self,
        freq: FrequencyOrWavelength,
        pixels: Pixels,
        nside: int,
        obs_time: Time,
        obs: str = "earth",
        obs_pos: u.Quantity[u.AU] | None = None,
        weights: u.Quantity[u.MJy / u.sr] | None = None,
        return_comps: bool = False,
        coord_in: Literal["E", "G", "C"] = "E",
    ) -> u.Quantity[u.MJy / u.sr]:
        """Returns the simulated binned zodiacal Emission given pixel numbers.

        The pixel numbers represent the pixel indicies on a HEALPix grid with resolution
        given by `nside`. The emission is binned to a HEALPix map with resolution given by
        `nside`.

        Args:
            freq: Delta frequency/wavelength or a sequence of frequencies corresponding to
                a bandpass over which to evaluate the zodiacal emission.
            pixels: HEALPix pixel indicies representing points on the celestial sphere.
            nside: HEALPix resolution parameter of the pixels and the binned map.
            obs_time: Time of observation.
            obs: Name of the Solar System observer. A list of all support observers (for a
                given ephemeridis) is specified in `supported_observers` attribute of the
                `Zodipy` instance. Defaults to 'earth'.
            obs_pos: The heliocentric ecliptic cartesian position of the observer in AU.
                Overrides the `obs` argument. Default is None.
            weights: Bandpass weights corresponding the the frequencies in `freq`. The
                weights are assumed to be in units of MJy/sr.
            return_comps: If True, the emission is returned component-wise. Defaults to False.
            coord_in: Coordinate frame of the input pointing. Assumes 'E' (ecliptic
                coordinates) by default.

        Returns:
            emission: Simulated zodiacal emission in units of 'MJy/sr'.

        """

        pixels = validate_pixels(pixels=pixels, nside=nside)

        unique_pixels, counts = np.unique(pixels, return_counts=True)
        unit_vectors = get_unit_vectors_from_pixels(
            coord_in=coord_in,
            pixels=unique_pixels,
            nside=nside,
        )

        return self._compute_emission(
            freq=freq,
            weights=weights,
            obs=obs,
            obs_time=obs_time,
            obs_pos=obs_pos,
            unit_vectors=unit_vectors,
            indicies=counts,
            binned=True,
            pixels=unique_pixels,
            nside=nside,
            return_comps=return_comps,
        )

    def _compute_emission(
        self,
        freq: FrequencyOrWavelength,
        weights: u.Quantity[u.MJy / u.sr] | None,
        obs: str,
        obs_time: Time,
        unit_vectors: npt.NDArray[np.float64],
        indicies: npt.NDArray[np.int64],
        binned: bool = False,
        obs_pos: u.Quantity[u.AU] | None = None,
        pixels: npt.NDArray[np.int64] | None = None,
        nside: int | None = None,
        return_comps: bool = False,
    ) -> u.Quantity[u.MJy / u.sr]:
        """Computes the component-wise zodiacal emission."""

        if not self.extrapolate:
            validate_frequency_in_model_range(freq=freq, model=self.ipd_model)

        normalized_weights = validate_and_normalize_weights(freq=freq, weights=weights)

        interpolated_source_params = interpolate_source_parameters(
            model=self.ipd_model, freq=freq, weights=normalized_weights
        )

        observer_position, earth_position = get_obs_and_earth_positions(
            obs=obs, obs_time=obs_time, obs_pos=obs_pos
        )

        start, stop = get_line_of_sight_endpoints(
            cutoff=self.los_dist_cut.value,
            obs_pos=observer_position,
            unit_vectors=unit_vectors,
        )

        emission = np.zeros(
            (self.ipd_model.n_comps, hp.nside2npix(nside) if binned else indicies.size)
        )

        # Convert to Hz if `freq` is in units of wavelength
        freq = freq.to(u.Hz, u.spectral())
        # Renormalize weights with respect to the spectrum in Hz.
        if normalized_weights is not None:
            normalized_weights /= np.trapz(normalized_weights, freq.value)

        # Some components require additional non-static parameters to be computed, such as
        # the earth-trailing feature which needs the earth position in addition to the
        # model parameters.
        computed_parameters = {
            "X_earth": np.expand_dims(earth_position, axis=-1),
        }
        # Construct component density functions
        density_partials = construct_density_partials(
            comps=list(self.ipd_model.comps.values()),
            computed_params=computed_parameters,
        )
        # Create partial function with fixed arguments. This partial will again be used
        # as the basis for another partial depnding on wether or not the code is run in
        # parallel.
        partial_common_integrand = partial(
            _get_emission_at_step,
            start=start,
            gauss_quad_degree=self.gauss_quad_degree,
            X_obs=np.expand_dims(observer_position, axis=-1),
            density_partials=density_partials,
            freq=freq.value,
            weights=normalized_weights,
            T_0=self.ipd_model.T_0,
            delta=self.ipd_model.delta,
            **asdict(interpolated_source_params),
        )
        # Distribute pointing to available CPUs and compute the emission in parallel.
        if self.parallel:
            n_proc = multiprocessing.cpu_count() if self.n_proc is None else self.n_proc

            unit_vector_chunks = np.array_split(unit_vectors, n_proc, axis=-1)
            stop_chunks = np.array_split(stop, n_proc, axis=-1)

            with multiprocessing.get_context(SYS_PROC_START_METHOD).Pool(
                processes=n_proc
            ) as pool:
                # Create a partial functions that will be passed to the quadpy integration
                # scheme. Arrays are reshaped to (d, n, p) for broadcasting purposes where
                # d is the geometrical dimensionality, n is the number of different
                # pointings, and p is the number of integration points of the quadrature.
                partial_integrand_chunks = [
                    partial(
                        partial_common_integrand,
                        stop=np.expand_dims(stop, axis=-1),
                        u_los=np.expand_dims(unit_vectors, axis=-1),
                    )
                    for unit_vectors, stop in zip(unit_vector_chunks, stop_chunks)
                ]

                proc_chunks = [
                    pool.apply_async(
                        self._integration_scheme.integrate,
                        args=(partial_integrand, [-1, 1]),
                    )
                    for partial_integrand in partial_integrand_chunks
                ]
                integrated_comp_emission = np.concatenate(
                    [result.get() for result in proc_chunks], axis=1
                )

        # Compute the emission only in the main process.
        else:
            unit_vectors_expanded = np.expand_dims(unit_vectors, axis=-1)
            stop_expanded = np.expand_dims(stop, axis=-1)

            partial_integrand = partial(
                partial_common_integrand,
                stop=stop_expanded,
                u_los=unit_vectors_expanded,
            )

            integrated_comp_emission = self._integration_scheme.integrate(
                partial_integrand, [-1, 1]
            )

        # We convert the integral from [-1, 1] back to [start, stop].
        integrated_comp_emission *= 0.5 * (stop - start)

        if binned:
            emission[:, pixels] = integrated_comp_emission
        else:
            emission = integrated_comp_emission[:, indicies]

        if self.solar_cut is not None:
            # The observer position is aquired in geocentric coordinates before being
            # rotated to ecliptic coordinates. This means that we can find the
            # heliocentric origin in this coordinate system by simply taking the negative
            # of the observer position. The emission corresponding to unit_vectors with a
            # angular distance (between the pointing and the sun) smaller than the
            # specified `solar_cutoff` value is masked with the `solar_cutoff_fill_value`.
            ang_dist = hp.rotator.angdist(-observer_position, unit_vectors)
            solar_mask = ang_dist < self.solar_cut.value
            if binned and pixels is not None:
                emission[:, pixels[solar_mask]] = self.solar_cut_fill_value
            else:
                emission[:, solar_mask[indicies]] = self.solar_cut_fill_value

        emission = (emission << SPECIFIC_INTENSITY_UNITS).to(u.MJy / u.sr)

        return emission if return_comps else emission.sum(axis=0)

    def __repr__(self) -> str:
        repr_str = f"{self.__class__.__name__}("
        for attribute_name, attribute in self.__dict__.items():
            if attribute_name.startswith("_"):
                continue
            repr_str += f"{attribute_name}={attribute!r}, "

        return repr_str[:-2] + ")"


def _get_emission_at_step(
    r: float | npt.NDArray[np.float64],
    start: float,
    stop: float | npt.NDArray[np.float64],
    gauss_quad_degree: int,
    X_obs: npt.NDArray[np.float64],
    u_los: npt.NDArray[np.float64],
    density_partials: tuple[PartialComputeDensityFunc],
    freq: float | npt.NDArray[np.float64],
    weights: npt.NDArray[np.float64] | None,
    T_0: float,
    delta: float,
    emissivities: npt.NDArray[np.float64],
    albedos: npt.NDArray[np.float64],
    phase_coefficients: tuple[float, ...],
    solar_irradiance: float,
) -> npt.NDArray[np.float64]:
    """Returns the zodiacal emission at a step along all lines of sight."""

    # Convert the line of sight range from [-1, 1] to the true ecliptic positions
    R_los = ((stop - start) / 2) * r + (stop + start) / 2

    X_los = R_los * u_los
    X_helio = X_los + X_obs
    R_helio = np.sqrt(X_helio[0] ** 2 + X_helio[1] ** 2 + X_helio[2] ** 2)

    temperature = get_dust_grain_temperature(R_helio, T_0, delta)

    if weights is not None:
        blackbody_emission = get_bandpass_integrated_blackbody_emission(
            freq=freq,
            weights=weights,
            T=temperature,
        )
    else:
        blackbody_emission = get_blackbody_emission(freq=freq, T=temperature)

    emission = np.zeros(
        (len(density_partials), np.shape(X_helio)[1], gauss_quad_degree)
    )
    density = np.zeros_like(emission)
    for idx, (get_density_func, albedo, emissivity) in enumerate(
        zip(density_partials, albedos, emissivities)
    ):
        density[idx] = get_density_func(X_helio)
        emission[idx] = (1 - albedo) * (emissivity * blackbody_emission)

    if any(albedo != 0 for albedo in albedos):
        solar_flux = solar_irradiance / R_helio**2
        scattering_angle = get_scattering_angle(R_los, R_helio, X_los, X_helio)
        phase_function = get_phase_function(scattering_angle, phase_coefficients)

        for idx, albedo in enumerate(albedos):
            emission[idx] += albedo * solar_flux * phase_function

    return emission * density



from typing import Sequence, TypeVar, Union

import astropy.units as u
import numpy as np
import numpy.typing as npt

Pixels = Union[int, Sequence[int], npt.NDArray[np.integer]]
SkyAngles = Union[u.Quantity[u.deg], u.Quantity[u.rad]]
FrequencyOrWavelength = Union[u.Quantity[u.Hz], u.Quantity[u.m]]
FloatOrNDArray = TypeVar("FloatOrNDArray", float, npt.NDArray[np.float64])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zodipy']

package_data = \
{'': ['*']}

install_requires = \
['astropy>=5.0.1',
 'healpy>=1.15.0,<2.0.0',
 'jplephem>=2.17,<3.0',
 'numba>=0.56.2,<0.57.0',
 'numpy>=1.22.3,<2.0.0',
 'quadpy>=0.16.13,<0.17.0',
 'typing-extensions>=4.1.1,<5.0.0']

setup_kwargs = {
    'name': 'zodipy',
    'version': '0.7.1',
    'description': 'Software for simulating zodiacal emission',
    'long_description': '\n<img src="docs/img/zodipy_logo.png" width="350">\n\n[![PyPI version](https://badge.fury.io/py/zodipy.svg)](https://badge.fury.io/py/zodipy)\n[![astropy](http://img.shields.io/badge/powered%20by-AstroPy-orange.svg?style=flat)](http://www.astropy.org/)\n![Tests](https://github.com/MetinSa/zodipy/actions/workflows/tests.yml/badge.svg)\n[![codecov](https://codecov.io/gh/Cosmoglobe/zodipy/branch/main/graph/badge.svg?token=VZP9L79EUJ)](https://codecov.io/gh/Cosmoglobe/zodipy)\n[![arXiv Paper](https://img.shields.io/badge/arXiv-2205.12962-green)](https://arxiv.org/abs/2205.12962)\n---\n\n\nZodiPy is a Python tool for simulating the zodiacal emission in intensity that an arbitrary Solar System observer sees, either in the form of timestreams or full-sky HEALPix maps.\n\n![plot](docs/img/zodipy_map.png)\n\n# Help\nSee the [documentation](https://cosmoglobe.github.io/zodipy/) for more information and examples on how to use ZodiPy for different applications.\n\n# Installation\nZodiPy is installed using `pip install zodipy`.\n\n# A simple example\n```python\nimport astropy.units as u\nfrom astropy.time import Time\n\nfrom zodipy import Zodipy\n\n\nmodel = Zodipy("dirbe")\n\nemission = model.get_emission_ang(\n    25 * u.micron,\n    theta=[10, 10.1, 10.2] * u.deg,\n    phi=[90, 89, 88] * u.deg,\n    obs_time=Time("2022-01-01 12:00:00"),\n    obs="earth",\n)\n\nprint(emission)\n#> [15.35392831 15.35495051 15.35616009] MJy / sr\n```\n\n# Scientific paper and citation\nFor an overview of the ZodiPy model approach and other information regarding zodiacal emission and interplanetary dust modeling we refer to the scientific paper on ZodiPy:\n- [Cosmoglobe: Simulating zodiacal emission with ZodiPy (San et al. 2022)](https://arxiv.org/abs/2205.12962). \n\nSee [CITATION](https://github.com/Cosmoglobe/zodipy/blob/dev/CITATION.bib) if you have used ZodiPy in your work and want to cite the software.\n\n# Funding\nThis work has received funding from the European Union\'s Horizon 2020 research and innovation programme under grant agreements No 776282 (COMPET-4; BeyondPlanck), 772253 (ERC; bits2cosmology) and 819478 (ERC; Cosmoglobe).\n\n\n<div style="display: flex; flex-direction: row; justify-content: space-evenly">\n    <img style="width: 49%; height: auto; max-width: 500px; align-self: center" src="https://user-images.githubusercontent.com/28634670/170697040-d5ec2935-29d0-4847-8999-9bc4eaa59e56.jpeg"> \n    &nbsp; \n    <img style="width: 49%; height: auto; max-width: 500px; align-self: center" src="https://user-images.githubusercontent.com/28634670/170697140-b010aa69-9f9a-44c0-b702-8a05ec0b6d3e.jpeg">\n</div>',
    'author': 'Metin San',
    'author_email': 'metinisan@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/MetinSa/zodipy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)



"""Functions that extract the positions of Solar System bodies using the astropy
`solar_system_ephemeris` API.

The lagrange point SEMB-L2 is not included in any of the current available
ephemerides. We implement an approximation to its position, assuming that 
SEMB-L2 is at all times located at a fixed distance from Earth along the vector 
pointing to Earth from the Sun.
"""

from __future__ import annotations

import astropy.units as u
import numpy as np
import numpy.typing as npt
from astropy.coordinates import HeliocentricMeanEcliptic, get_body
from astropy.time import Time

DISTANCE_FROM_EARTH_TO_L2 = u.Quantity(0.009896235034000056, u.AU)
DISTANCE_TO_JUPITER = u.Quantity(5.2, u.AU)


def get_obs_and_earth_positions(
    obs: str, obs_time: Time, obs_pos: u.Quantity[u.AU] | None
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:

    earth_position = _get_earth_position(obs_time)
    if obs_pos is None:
        obs_pos = _get_observer_position(obs, obs_time, earth_position)

    return obs_pos.reshape(3, 1).value, earth_position.reshape(3, 1).value


def _get_earth_position(obs_time: Time) -> u.Quantity[u.AU]:
    """Returns the position of the Earth given an ephemeris and observation time."""

    earth_skycoordinate = get_body("earth", obs_time)
    earth_skycoordinate = earth_skycoordinate.transform_to(HeliocentricMeanEcliptic)

    return earth_skycoordinate.cartesian.xyz.to(u.AU)


def _get_observer_position(
    obs: str, obs_time: Time, earth_pos: u.Quantity[u.AU]
) -> u.Quantity[u.AU]:
    """Returns the position of the Earth and the observer."""

    if obs.lower() == "semb-l2":
        return _get_sun_earth_moon_barycenter(earth_pos)

    observer_skycoordinate = get_body(obs, obs_time)
    observer_skycoordinate = observer_skycoordinate.transform_to(
        HeliocentricMeanEcliptic
    )

    return observer_skycoordinate.cartesian.xyz.to(u.AU)


def _get_sun_earth_moon_barycenter(
    earth_position: u.Quantity[u.AU],
) -> u.Quantity[u.AU]:
    """Returns the approximated position of SEMB-L2 given Earth's position."""

    r_earth = np.linalg.norm(earth_position)

    earth_unit_vector = earth_position / r_earth
    semb_l2_length = r_earth + DISTANCE_FROM_EARTH_TO_L2

    return earth_unit_vector * semb_l2_length



from __future__ import annotations

import inspect
from dataclasses import asdict
from functools import partial
from typing import Any, Callable, Sequence

import numpy as np
import numpy.typing as npt

from ._ipd_comps import Band, Cloud, Component, Feature, Ring

"""The density functions for the different types of components. Common for all of these 
is that the first argument will be `X_helio` (the line of sight from the observer towards
a point on the sky) and the remaining arguments will be parameters that are set by the
`Component` subclasses. These arguments are unpacked automatically, so for a 
ComputeDensityFunc to work, the mapped `Component` class must have all the parameters as 
instance variables. 
"""
ComputeDensityFunc = Callable[..., npt.NDArray[np.float64]]


def compute_cloud_density(
    X_helio: npt.NDArray[np.float64],
    X_0: npt.NDArray[np.float64],
    sin_Omega_rad: float,
    cos_Omega_rad: float,
    sin_i_rad: float,
    cos_i_rad: float,
    n_0: float,
    mu: float,
    alpha: float,
    beta: float,
    gamma: float,
) -> npt.NDArray[np.float64]:
    """
    Returns the dust density of a component at points in the Solar System
    given by 'X_helio' and the parameters of the component.
    """

    X_cloud = X_helio - X_0
    R_cloud = np.sqrt(X_cloud[0] ** 2 + X_cloud[1] ** 2 + X_cloud[2] ** 2)

    Z_cloud = (
        X_cloud[0] * sin_Omega_rad * sin_i_rad
        - X_cloud[1] * cos_Omega_rad * sin_i_rad
        + X_cloud[2] * cos_i_rad
    )

    ζ = np.abs(Z_cloud / R_cloud)

    g = np.where(ζ < mu, ζ**2 / (2 * mu), ζ - (mu / 2))

    return n_0 * R_cloud**-alpha * np.exp(-beta * g**gamma)


def compute_band_density(
    X_helio: npt.NDArray[np.float64],
    X_0: npt.NDArray[np.float64],
    sin_Omega_rad: float,
    cos_Omega_rad: float,
    sin_i_rad: float,
    cos_i_rad: float,
    n_0: float,
    delta_zeta_rad: float,
    p: float,
    v: float,
    delta_r: float,
) -> npt.NDArray[np.float64]:
    """
    Returns the dust density of a component at points in the Solar System
    given by 'X_helio' and the parameters of the component.
    """

    X_band = X_helio - X_0
    R_band = np.sqrt(X_band[0] ** 2 + X_band[1] ** 2 + X_band[2] ** 2)

    Z_band = (
        X_band[0] * sin_Omega_rad * sin_i_rad
        - X_band[1] * cos_Omega_rad * sin_i_rad
        + X_band[2] * cos_i_rad
    )

    ζ = np.abs(Z_band / R_band)
    ζ_over_δ_ζ = ζ / delta_zeta_rad
    term1 = 3 * n_0 / R_band
    term2 = np.exp(-(ζ_over_δ_ζ**6))

    # Differs from eq 8 in K98 by a factor of 1/self.v. See Planck XIV
    # section 4.1.2.
    term3 = 1 + (ζ_over_δ_ζ**p) / v

    term4 = 1 - np.exp(-((R_band / delta_r) ** 20))

    return term1 * term2 * term3 * term4


def compute_ring_density(
    X_helio: npt.NDArray[np.float64],
    X_0: npt.NDArray[np.float64],
    sin_Omega_rad: float,
    cos_Omega_rad: float,
    sin_i_rad: float,
    cos_i_rad: float,
    n_0: float,
    R: float,
    sigma_r: float,
    sigma_z: float,
) -> npt.NDArray[np.float64]:
    """
    Returns the dust density of a component at points in the Solar System
    given by 'X_helio' and the parameters of the component.
    """

    X_ring = X_helio - X_0
    R_ring = np.sqrt(X_ring[0] ** 2 + X_ring[1] ** 2 + X_ring[2] ** 2)

    Z_ring = (
        X_ring[0] * sin_Omega_rad * sin_i_rad
        - X_ring[1] * cos_Omega_rad * sin_i_rad
        + X_ring[2] * cos_i_rad
    )
    # Differs from eq 9 in K98 by a factor of 1/2 in the first and last
    # term. See Planck 2013 XIV, section 4.1.3.
    term1 = -((R_ring - R) ** 2) / sigma_r**2
    term2 = np.abs(Z_ring) / sigma_z

    return n_0 * np.exp(term1 - term2)


def compute_feature_density(
    X_helio: npt.NDArray[np.float64],
    X_0: npt.NDArray[np.float64],
    X_earth: npt.NDArray[np.float64],
    sin_Omega_rad: float,
    cos_Omega_rad: float,
    sin_i_rad: float,
    cos_i_rad: float,
    n_0: float,
    R: float,
    sigma_r: float,
    sigma_z: float,
    theta_rad: float,
    sigma_theta_rad: float,
) -> npt.NDArray[np.float64]:
    """
    Returns the dust density of a component at points in the Solar System
    given by 'X_helio' and the parameters of the component.
    """

    X_feature = X_helio - X_0
    R_feature = np.sqrt(X_feature[0] ** 2 + X_feature[1] ** 2 + X_feature[2] ** 2)

    Z_feature = (
        X_feature[0] * sin_Omega_rad * sin_i_rad
        - X_feature[1] * cos_Omega_rad * sin_i_rad
        + X_feature[2] * cos_i_rad
    )
    X_earth_comp = X_earth - X_0

    theta_comp = np.arctan2(X_feature[1], X_feature[0]) - np.arctan2(
        X_earth_comp[1], X_earth_comp[0]
    )

    delta_theta = theta_comp - theta_rad
    delta_theta = np.where(delta_theta < -np.pi, +2 * np.pi, delta_theta)
    delta_theta = np.where(delta_theta > np.pi, -2 * np.pi, delta_theta)

    # Differs from eq 9 in K98 by a factor of 1/2 in the first and last
    # term. See Planck 2013 XIV, section 4.1.3.
    exp_term = (R_feature - R) ** 2 / sigma_r**2
    exp_term += np.abs(Z_feature) / sigma_z
    exp_term += delta_theta**2 / sigma_theta_rad**2

    return n_0 * np.exp(-exp_term)


# Mapping of implemented zodiacal component data classes and their density functions.
DENSITY_FUNCS: dict[type[Component], ComputeDensityFunc] = {
    Cloud: compute_cloud_density,
    Band: compute_band_density,
    Ring: compute_ring_density,
    Feature: compute_feature_density,
}

PartialComputeDensityFunc = Callable[[npt.NDArray[np.float64]], npt.NDArray[np.float64]]


def construct_density_partials(
    comps: Sequence[Component],
    computed_params: dict[str, Any],
) -> tuple[PartialComputeDensityFunc, ...]:
    """
    Construct the density functions for the components.
    Raises exception for incorrectly defined components or component density functions.
    """

    partial_density_funcs: list[PartialComputeDensityFunc] = []
    for comp in comps:
        comp_dict = asdict(comp)
        func_params = inspect.signature(DENSITY_FUNCS[type(comp)]).parameters.keys()
        residual_params = [key for key in func_params if key not in comp_dict.keys()]
        try:
            residual_params.remove("X_helio")
        except ValueError:
            raise ValueError(
                "X_helio must be be the first argument to the density function of a component."
            )

        if residual_params:
            if residual_params - computed_params.keys():
                raise ValueError(
                    f"Argument(s) {residual_params} required by the density function "
                    f"{DENSITY_FUNCS[type(comp)]} are not provided by instance variables in "
                    f"{type(comp)} or by the `computed_parameters` dict."
                )
            comp_dict.update(computed_params)

        # Remove excess intermediate parameters from the component dict.
        comp_params = {
            key: value for key, value in comp_dict.items() if key in func_params
        }

        partial_func = partial(DENSITY_FUNCS[type(comp)], **comp_params)
        partial_density_funcs.append(partial_func)

    return tuple(partial_density_funcs)



from __future__ import annotations

import numpy as np
import numpy.typing as npt

EPS = float(np.finfo(float).eps)


def get_line_of_sight_endpoints(
    cutoff: float,
    obs_pos: npt.NDArray[np.float64],
    unit_vectors: npt.NDArray[np.float64],
) -> tuple[float, npt.NDArray[np.float64]]:
    """Returns the start and stop positions along the line of sights."""

    x, y, z = obs_pos.flatten()

    if cutoff < (r := np.sqrt(x**2 + y**2 + z**2)):
        raise ValueError(f"los_dist_cut is {cutoff} but observer_pos is {r}")

    u_x, u_y, u_z = unit_vectors
    lon = np.arctan2(u_y, u_x)
    lat = np.arcsin(u_z)

    cos_lat = np.cos(lat)
    b = 2 * (x * cos_lat * np.cos(lon) + y * cos_lat * np.sin(lon))
    c = r**2 - cutoff**2

    q = -0.5 * b * (1 + np.sqrt(b**2 - 4 * c) / np.abs(b))

    return EPS, np.maximum(q, c / q)
